//
//  File.swift
//  
//
//  Created by Malcolm Anderson on 6/9/21.
//

import Foundation

public struct SpeedAtPosition: RMCCommand {
    public var id = 36
    
    public var position: Double
    public var velocity: Double
    
    public init(toSpeed speed: Double, atPosition position: Double) {
        self.position = position
        self.velocity = speed
    }
    
    public var argArray: [Double] {[self.position, self.velocity]}
}

public struct SineStart: RMCCommand {
    public var id = 72
    
    public var offset: Double
    public var amplitude: Double
    public var frequency: Double
    public var cycles: Double
    public var startLocation: SineStartLocation
    public var statusBlock: DMCPAddress?
    
    public init(withOffset offset: Double, amplitude: Double, frequency: Double, cycles: Double = 0, startLocation: SineStartLocation = .auto, statusBlock: DMCPAddress? = nil) {
        self.offset = offset
        self.amplitude = amplitude
        self.frequency = frequency
        self.cycles = cycles
        self.startLocation = startLocation
        self.statusBlock = statusBlock
    }
    
    public var argArray: [Double] {[self.offset, self.amplitude, self.frequency, self.cycles, self.startLocation.rawValue, self.statusBlock != nil ? Double(self.statusBlock!.intValue) : 0.0]}
}

public struct SineStop: RMCCommand {
    public var id = 73
    
    public var stopLocation: SineStopLocation
    
    public init(at location: SineStopLocation) {
        self.stopLocation = location
    }
    
    public var argArray: [Double] {[self.stopLocation.rawValue]}
}

public struct ChangeMaster: RMCCommand {
    public var id = 79
    
    public var masterRegister: DMCPAddress
    
    public init(to register: DMCPAddress) {
        self.masterRegister = register
    }
    
    public var argArray: [Double] {[Double(self.masterRegister.intValue)]}
}

public struct ChangeTargetParameter: RMCCommand {
    public var id = 80
    
    public var parameter: CommandParameter
    public var newValue: Double
    public var rampType: RampType
    public var rampValue: Double
    
    public init(_ parameter: CommandParameter, toValue newValue: Double, withRampType rampType: RampType, toRampValue rampValue: Double) {
        self.parameter = parameter
        self.newValue = newValue
        self.rampType = rampType
        self.rampValue = rampValue
    }
    
    public var argArray: [Double] {[self.parameter.rawValue, self.newValue, self.rampType.rawValue, self.rampValue]}
}

public struct CurveAdd: RMCCommand {
    public var id = 82
    
    public var curveID: Int
    public var dataAddress: DMCPAddress
    public var interpolationMethod: CurveInterpolationMethod
    public var lifeCycle: CurveLifeCycle
    
    public init(withID id: Int, at address: DMCPAddress, withInterpolation interp: CurveInterpolationMethod, withLifeCycle lifeCycle: CurveLifeCycle) {
        self.curveID = id
        self.dataAddress = address
        self.interpolationMethod = interp
        self.lifeCycle = lifeCycle
    }
    
    public var argArray: [Double] {[Double(self.curveID), Double(self.dataAddress.intValue), self.interpolationMethod.rawValue, self.lifeCycle.rawValue]}
}

public struct CurveDelete: RMCCommand {
    public var id = 83
    
    public var curveID: Int
    public var count: Int
    
    public init(startingAtID id: Int, count: Int) {
        self.curveID = id
        self.count = count
    }
    
    public var argArray: [Double] {[Double(self.curveID), Double(self.count)]}
}

public struct CurveDeleteExcept: RMCCommand {
    public var id = 84
    
    public var curveID: Int
    public var count: Int
    
    public init(startingAtID id: Int, count: Int) {
        self.curveID = id
        self.count = count
    }
    
    public var argArray: [Double] {[Double(self.curveID), Double(self.count)]}
}

public struct CurveDeleteAll: RMCCommand {
    public var id = 85
    
    public var argArray: [Double] {[]}
}

public struct CurveStart: RMCCommand {
    public var id = 86
    
    public var curveID: Int
    public var masterRegister: DMCPAddress
    public var cycles: Double
    
    public init(withID id: Int, on register: DMCPAddress, forCycles cycles: Double = 0.0) {
        self.curveID = id
        self.masterRegister = register
        self.cycles = cycles
    }
    
    public var argArray: [Double] {[Double(self.curveID), Double(self.masterRegister.intValue), self.cycles]}
}

public struct CurveStartAdvanced: RMCCommand {
    public var id = 88
    
    public var curveID: Int
    public var masterRegister: DMCPAddress
    public var cycles: Double
    public var curveAlignment: CurveAlignment
    public var masterAlignment: MasterAlignment
    
    public var curveScale: Double
    public var curveOffset: Double
    public var masterScale: Double
    public var masterOffset: Double
    public var statusBlock: DMCPAddress?
    
    public init(withID id: Int, on register: DMCPAddress, forCycles cycles: Double, withCurveAlignment curveAlignment: CurveAlignment, withMasterAlignment masterAlignment: MasterAlignment, curveScale: Double = 1.0, curveOffset: Double = 0.0, masterScale: Double = 1.0, masterOffset: Double = 0.0, statusBlock: DMCPAddress? = nil) {
        self.curveID = id
        self.masterRegister = register
        self.cycles = cycles
        self.curveAlignment = curveAlignment
        self.masterAlignment = masterAlignment
        self.curveScale = curveScale
        self.curveOffset = curveOffset
        self.masterScale = masterScale
        self.masterOffset = masterOffset
        self.statusBlock = statusBlock
    }
    
    public var argArray: [Double] {
        var options: Double = 0.0
        options += self.curveAlignment.rawValue
        options += self.masterAlignment.rawValue
        switch self.masterAlignment {
        case .absolute(let alignment):
            options += alignment.rawValue
        case .relative(let alignment):
            options += alignment.rawValue
        }
        
        return [Double(self.curveID), Double(self.masterRegister.intValue), self.cycles, options, self.curveScale, self.curveOffset, self.masterScale, self.masterOffset, self.statusBlock != nil ? Double(self.statusBlock!.intValue) : 0.0]
    }
}
