//
//  File.swift
//  
//
//  Created by Malcolm Anderson on 6/15/21.
//

import Foundation

public protocol DMCPRequest {
    var file: UInt16 { get set }
    var element: UInt16 { get set }
    
    func data(transactionID: UInt16) -> Data
}

/**
 Stores the information for a write request to an RMC.
 - Parameters:
    - file: The starting address's file part
    - element: The starting address's element part
    - associatedData: Data that may uniquely identify this request, if desired
    - values: The values to write to the RMC's memory, starting at the address determined by `file` and `element`
    - onResult: A closure to execute once there is a result from this request. If the request was successful, `error` will be `nil`; if the request failed, `error` will describe why.
 */
public struct DMCPWriteRequest: DMCPRequest {
    public var file: UInt16
    public var element: UInt16
    public var values: [DMCPData] = []
    public var onResult: (_ error: Error?) -> Void = {_ in }
    
    public init(_ file: UInt16, _ element: UInt16, values: [DMCPData] = [], onResult: @escaping (_ error: Error?) -> Void = {_ in }) {
        self.file = file
        self.element = element
        self.values = values
        self.onResult = onResult
    }
    
    public init(_ file: UInt16, _ element: UInt16, value: DMCPData, onResult: @escaping (_ error: Error?) -> Void = {_ in}) {
        self.init(file, element, values: [value], onResult: onResult)
    }
    
    public func data(transactionID: UInt16) -> Data {
        var data = Data()
        
        // Packet length
        data.append(UInt16(14 + 4 * values.count))

        // Static values
        data.append(UInt16(0x0200))

        // Transaction ID
        data.append(transactionID)

        // Function code
        data.append(UInt8(0x15))

        // Byte order (little endian)
        data.append(UInt8(0x00))
        
        // Starting address (file)
        data.append(file)

        // Starting address (element)
        data.append(element)

        // Write count (less than 1024)
        data.append(UInt16(values.count))

        // Reserved
        data.append(UInt16(0))

        // Data
        for value in values {
            data.append(value)
        }
        
        return data
    }
}

/**
 Stores the information for a read request to an RMC.
 - Parameters:
    - file: The starting address's file part
    - element: The starting address's element part
    - readCount: The number of variables to read the data from, starting at the address determined by `file` and `element`
    - associatedData: Data that may uniquely identify this request, if desired
    - onResult: A closure to execute once there is a result from this request. If the request was successful, `error` will be `nil`; if the request failed, `error` will describe why. The parameter `values` will contain a list of the retrieved values if no error occured, and will be empty otherwise.
 */
public struct DMCPReadRequest: DMCPRequest {
    public var file: UInt16
    public var element: UInt16
    public var readTypes: [DMCPData.DataType] = []
    public var onResult: (_ error: Error?, _ values: [DMCPData]) -> Void = {_,_ in }
    
    public init(_ file: UInt16, _ element: UInt16, readTypes: [DMCPData.DataType] = [], onResult: @escaping (_ error: Error?, _ value: [DMCPData]) -> Void = {_,_ in }) {
        self.file = file
        self.element = element
        self.readTypes = readTypes
        self.onResult = onResult
    }
    
    public init(_ file: UInt16, _ element: UInt16, readType: DMCPData.DataType, onResult: @escaping (_ error: Error?, _ value: DMCPData?) -> Void = {_,_ in}) {
        self.file = file
        self.element = element
        self.readTypes = [readType]
        self.onResult = { error, value in
            onResult(error, value.isEmpty ? nil : value[0])
        }
    }
    
    public init(_ file: UInt16, _ element: UInt16, intCount: Int = 1, onResult: @escaping (_ error: Error?, _ value: Int32?) -> Void = {_,_ in }) {
        self.init(file, element, readType: .int) { error, value in
            if case .int(let val) = value {
                onResult(error, val)
            }
            else {
                onResult(error, nil)
            }
        }
    }
    
    public init(_ file: UInt16, _ element: UInt16, realCount: Int = 1, onResult: @escaping (_ error: Error?, _ value: Float32?) -> Void = {_,_ in }) {
        self.init(file, element, readType: .real) { error, value in
            if case .real(let val) = value {
                onResult(error, val)
            }
            else {
                onResult(error, nil)
            }
        }
    }
    
    public func data(transactionID: UInt16) -> Data {
        var data = Data()
        
        // Packet length
        data.append(UInt16(0x000C))
        
        // Static values
        data.append(UInt16(0x0200))
        
        // Transaction ID
        data.append(transactionID)
        
        // Function code
        data.append(UInt8(0x14))
        
        // Byte order (little endian)
        data.append(UInt8(0x0))
        
        // Starting address (file)
        data.append(file)
        
        // Starting address (element)
        data.append(element)
        
        // Read count
        data.append(UInt16(readTypes.count))
        
        return data
    }
}
