//
//  File.swift
//  
//
//  Created by Malcolm Anderson on 6/21/21.
//

import Foundation

public class AdvancedPlotReader: PlotReader {
    internal let connection: DMCPConnection
    
    public required init(_ connection: DMCPConnection) {
        self.connection = connection
    }
    
    public func readPlot(_ id: Int, waitTimeout: Double? = nil, onProgress: @escaping (Int, Int) -> Void = {_,_ in}, onResult: @escaping (Error?, RMCPlotData?) -> Void) {
        // (Same code as BasicPlotReader)
        // Determine RMC type
        guard let rmcType = self.connection.rmcType?.toRMCClass() else {
            onResult(nil, nil)
            return
        }
        
        // Set up variables
        let requestedReadSamples = 50
        var totalAddresses: [DMCPAddress] = []
        var totalData: [DMCPAddress: [Double?]] = [:]
        
        // Ensure plot state register is 2
        let startDate = Date()
        func readPlotState() {
            // We need to wait to read the plot until the plot state is complete
            self.connection.readReal(rmcType.plotStateAddr(id)!) { error, plotState in
                if error != nil {
                    print("Error getting plot state: \(error!)")
                    return
                }
                let plotStateAsEnum = PlotState(rawValue: Int(plotState!))
                
                if plotStateAsEnum == .complete {
                    readPlotTotalSamples()
                } else {
                    print("Plot state is \(String(describing: plotStateAsEnum))... we have to wait")
                    let timeSpan = startDate.timeIntervalSince(Date())
                    
                    if (waitTimeout != nil && waitTimeout! > timeSpan) {
                        onResult(RMCPlotError.timeout, nil)
                    } else {
                        readPlotState()
                    }
                }
            }
        }
        
        var capturedSamples = 0
        func readPlotTotalSamples() {
            // Read number of samples captured
            self.connection.readReal(rmcType.plotCapturedSamplesAddr(id)!) { error, samples in
                if error != nil {
                    print("Error getting captured samples: \(error!)")
                    onResult(error, nil)
                    return
                }
                
                capturedSamples = Int(samples!)
                
                getDataItems()
            }
        }
        
        /*
         STEP 2
         We need to find out how many data items we're capturing.
         To do this, we can use getPlotAddresses.
         */
        var totalItemsToRead = 0
        func getDataItems() {
            self.connection.getPlotAddresses(forPlot: id) { error, addresses in
                if error != nil {
                    print("Error getting plot address: \(error!)")
                    onResult(error, nil)
                    return
                }
                
                for address in addresses {
                    totalData[address] = []
                }
                
                totalAddresses = addresses
                totalItemsToRead = totalAddresses.count * capturedSamples
                writeRequestedReadSamples()
            }
        }
        
        /*
         STEP 3
         Next, we tell the RMC how many read samples we'd like for each
         data item at a time.
         */
        func writeRequestedReadSamples() {
            self.connection.writeInt(rmcType.plotDynamicRequestedReadSamplesAddr(forPlot: id)!, value: requestedReadSamples) { error in
                if error != nil {
                    print("Error writing requested read samples: \(error!)")
                    onResult(error, nil)
                    return
                }

                writeOneToUploadModeRegister()
            }
        }
        
        /*
         STEP 4
         To start actually reading data, we write 1 to the upload mode register.
         */
        func writeOneToUploadModeRegister() {
            self.connection.writeInt(rmcType.plotDynamicUploadModeStatusAddr(forPlot: id)!, value: 1) { error in
                if error != nil {
                    print("Error writing 1 to upload mode/status register: \(error!)")
                    onResult(error, nil)
                    return
                }
                
                // Next, read the number of samples uploaded
                readPlotData()
            }
        }
        
        /*
         STEP 5
         In order to read the data, we must first read the dynamic plot samples uploaded register, which
         seemingly triggers new data to be loaded into the data registers.
         */
        // Total samples to read should be the number of samples, times the number of items
        var readSoFar = 0
        func readPlotData() {
            // Read the number of samples uploaded per data item
            self.connection.readInt(rmcType.plotDynamicSamplesUploadedAddr(forPlot: id)!) { error, uploaded in
                if error != nil {
                    print("Error reading the number of samples uploaded: \(error!)")
                    onResult(error, nil)
                    return
                }
                
                guard let samplesUploaded = uploaded else {
                    print("Samples uploaded is nil - shouldn't happen")
                    return
                }
                

                self.connection.readReals(rmcType.plotDynamicDataAddr(forPlot: id)!, count: totalData.count * Int(samplesUploaded)) { error, data in
                    if error != nil {
                        print("Error reading data from dynamic data area: \(error!)")
                        onResult(error, nil)
                        return
                    }
                    
                    var startIndex = 0
                    for address in totalAddresses {
                        let endIndex = startIndex + Int(samplesUploaded)
                        
                        for i in startIndex..<endIndex {
                            totalData[address]?.append(data[i])
                        }
                        
                        startIndex += Int(samplesUploaded)
                        
                        readSoFar += Int(samplesUploaded)
                        onProgress(readSoFar, totalItemsToRead)
                    }
                    
                    // If we read fewer samples than we requested,
                    // we must now be done
                    if samplesUploaded < requestedReadSamples {
                        onProgress(1, 1)
                        PlotUtilities.packageData(totalData, forPlot: id, connection: self.connection, onResult)
                    }
                    
                    // Otherwise, there's probably more to go
                    else {
                        readPlotData()
                    }
                }
                
            }

        }
        
        readPlotState()
    }
    
}
