//
//  File.swift
//  
//
//  Created by Malcolm Anderson on 6/16/21.
//

import Foundation

/**
 Reads plots from an RMC using the "Basic" method described in
 [Reading Plots with a Host Controller](https://deltamotion.com/support/webhelp/rmctools/Controller_Features/Plots/Plots_Reading_with_a_Host_Controller.htm).
 */
public class BasicPlotReader: PlotReader {
    internal let connection: DMCPConnection
    
    /**
     - Parameter connection: Initializes this plot reader to use this connection for requests.
     */
    public required init(_ connection: DMCPConnection) {
        self.connection = connection
    }
    
    /**
     Attempts to read the given plot and return it as an `RMCPlotData` structure.
     - Parameters:
        - id: The number of the plot to get.
        - onResult: A closure to execute once there is a result from this request.
        - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
        - data: The plot data. If an error occurred, this is `nil`.
     */
    public func readPlot(_ id: Int, waitTimeout: Double? = nil, onProgress: @escaping (_ samplesRead: Int, _ totalSamples: Int) -> Void = {_,_  in}, onResult: @escaping (_ error: Error?, _ data: RMCPlotData?) -> Void) {
        
        // Determine RMC type
        guard let rmcType = self.connection.rmcType?.toRMCClass() else {
            // TODO: make an error for this
            onResult(nil, nil)
            return
            
        }

        // Ensure plot state register is 2
        let startDate = Date()
        func readPlotState() {
            // We need to wait to read the plot until the plot state is complete
            self.connection.readReal(rmcType.plotStateAddr(id)!) { error, plotState in
                if error != nil {
                    print("Error getting plot state: \(error!)")
                    return
                }
                let plotStateAsEnum = PlotState(rawValue: Int(plotState!))
                
                if plotStateAsEnum == .complete {
                    readPlotData()
                } else {
                    print("Plot state is \(String(describing: plotStateAsEnum))... we have to wait")
                    let timeSpan = startDate.timeIntervalSince(Date())
                    
                    if (waitTimeout != nil && waitTimeout! > timeSpan) {
                        onResult(RMCPlotError.timeout, nil)
                    } else {
                        readPlotState()
                    }
                }
            }
        }
        
        func readPlotData() {
            // Dictionary to hold registers and their samples
            var data: [DMCPAddress: [Double?]] = [:]
            
            // Keep track of total progress
            var currentDataSet = 0
            let maxDataSets = rmcType.staticPlotReadMaxDataSets
            
            
            // Read number of samples captured
            self.connection.readReal(rmcType.plotCapturedSamplesAddr(id)!) { error, samples in
                if error != nil {
                    print("Error getting captured samples: \(error!)")
                    onResult(error, nil)
                    return
                }
                
                var totalDataSoFar = 0
                let totalDataToRead = maxDataSets * Int(samples!)
                
                func readDataSet() {
                    // We know how many samples were captured, so we can continue
                    let endSamples = Int(samples ?? 0.0)
                    let maxSamplesToRead = 256
                    var readSoFar = 0
                    
                    // Approach: have a small function that takes an existing array,
                    // appends new values to it, then either calls itself again or
                    // returns the array (recursion)
                    
                    var totalSamples: [Double?] = []
                    
                    let baseAddress = rmcType.plotStaticUploadAreaAddr(forPlot: id, dataSet: currentDataSet)
                    
                    func readMoreSamples() {
                        // We want to read maxSamplesToRead, if possible.
                        var readThisTime = maxSamplesToRead
                        
                        // If there aren't that many samples to read, though,
                        // we want to read everything that's remaining.
                        if readThisTime > (endSamples - readSoFar) {
                            readThisTime = endSamples - readSoFar
                        }
                        
                        var address = baseAddress
                        address!.element += UInt16(readSoFar)
                        
                        readSoFar += readThisTime
                        
                        self.connection.readReals(address!, count: readThisTime) { error, values in
                            if error != nil {
                                print("Error reading samples at \(readSoFar) for plot \(id) dataset \(currentDataSet) (address \(address!)): \(error!)")
                                onResult(error, nil)
                                return
                            }
                            
                            // No error occurred. Append the array we just got to the existing one
                            totalSamples.append(contentsOf: values)
                            
                            // Update callback closure with data read so far
                            totalDataSoFar += readThisTime
                            onProgress(totalDataSoFar, totalDataToRead)
                            
                            // Recurse if we still have samples to read
                            if readSoFar < endSamples {
                                readMoreSamples()
                            } else {
                                
                                // We have all of the samples for this data set!
                                // Let's get the address this data corresponds to.
                                self.connection.getPlotAddress(forPlot: id, atIndex: currentDataSet) { error, address in
                                    if error != nil {
                                        print("Couldn't get address for plot \(id) at index \(currentDataSet)")
                                        onResult(error, nil)
                                        return
                                    }
                                    
                                    if (address?.file ?? 0) != 0 && (address?.element ?? 0) != 0 {
                                        data[address!] = totalSamples
                                    }
                                    
                                    if currentDataSet < maxDataSets - 1 {
                                        currentDataSet += 1
                                        readDataSet()
                                    } else {
                                        PlotUtilities.packageData(data, forPlot: id, connection: self.connection, onResult)
                                    }
                                }
                            }
                        }
                    }
                    readMoreSamples()
                }
                readDataSet()
            }
        }
        readPlotState()
    }
}

/**
 Packaged plot data from an RMC, suitable for reading and processing.
 */
public struct RMCPlotData: Encodable {
    
    /// The sample period (i.e. the number of seconds in between samples) of the plot.
    public let samplePeriod: Double
    
    /// An array of quantity items captured by the plot.
    public let quantities: [RMCPlotQuantity]
    
    /// A dictionary relating register addresses (as strings) to the values they captured, in order.
    public let samples: [String: [Double?]]
}

/**
 Information about a single register/quantity captured by an RMC's plot.
 */
public struct RMCPlotQuantity: Encodable {
    /// A human-readable name of the quantity (i.e. "Axis0 Actual Position").
    public let name: String
    
    /// The unit or unit type for the quantity.
    public let unit: String
    
    /// The scale group that the quantity belongs to.
    public let scaleGroup: String
    
    /// Whether or not to center the scale group on zero.
    public let centerOnZero: Bool
    
    /// The formula (i.e. address) for the quantity.
    public let formula: String
    
    /// The color assigned to the quantity for drawing in plots.
    public let color: RMCPlotColor
}

/**
 Error reading plot data
 */
public enum RMCPlotError: Error {
    /// The plot data was not ready in the amount of time provided
    case timeout
}
