//
//  File.swift
//  
//
//  Created by Malcolm Anderson on 6/17/21.
//

import Foundation

/// Contains basic RGB color information.
public struct RMCPlotColor: Encodable {
    /// The red component of this color.
    public var r: Double = 1.0
    
    /// The green component of this color.
    public var g: Double = 1.0
    
    /// The blue component of this color.
    public var b: Double = 1.0
    
    /// Initializes this color to white.
    public init() {}
    
    
    /**
    Initializes this color with the given red, green, and blue values.
     - Parameters:
        - red: The red component of this color.
        - green: The green component of this color.
        - blue: The blue component of this color.
    */
    public init(red: Double, green: Double, blue: Double) {
        self.r = red
        self.g = green
        self.b = blue
    }
    
    /**
     Initializes this color with the given hue, saturation, and value.
     - Parameters:
        - hue: The hue of this color.
        - saturation: The saturation of this color.
        - value: The value of this color.
     */
    public init(hue: Double, saturation: Double, value: Double) {
        // Using formulas from https://www.rapidtables.com/convert/color/hsv-to-rgb.html
        let hueDegrees = hue * 360.0
        
        let c = value * saturation
        
        let x = c * (1 - abs((hueDegrees / 60.0).truncatingRemainder(dividingBy: 2.0) - 1))
        
        let r_: Double, g_: Double, b_: Double
        
        switch hueDegrees {
        case 0..<60.0:
            (r_, g_, b_) = (c, x, 0)
        case 60..<120:
            (r_, g_, b_) = (x, c, 0)
        case 120..<180:
            (r_, g_, b_) = (0, c, x)
        case 180..<240:
            (r_, g_, b_) = (0, x, c)
        case 240..<300:
            (r_, g_, b_) = (x, 0, c)
        case 300..<360:
            (r_, g_, b_) = (c, 0, x)
        default:
            (r_, g_, b_) = (1, 0, 1)
        }
        
        self.r = r_
        self.g = g_
        self.b = b_
    }
    
    /**
     Initializes this color using a seed which determines the hue, a saturation, and value.
     - Parameters:
        - seed: An integer to use to determine this color's hue.
        - saturation: The saturation of this color.
        - value: The value of this color.
     */
    public init(seed: Int, saturation: Double = 1.0, value: Double = 0.9) {
        let hue = Double((0 + seed * 250) % 360) / 360.0
        self.init(hue: hue, saturation: saturation, value: value)
    }
}
