//
//  File.swift
//  
//
//  Created by Malcolm Anderson on 6/15/21.
//

import Foundation

extension RMC {
    
    public static func getDisplayName(_ element: Int, axisType: AxisType, primaryInputType: InputType, secondaryInputType: InputType) -> RMCRegisterInfo {
        // First, let's handle all of the cases where all axis types share the same output
        var potentialName = commonAxisStatusRegisterNames[element]
        if potentialName != nil {
            return potentialName!
        }
        
        // Check primary input type
        if 8 <= element && element <= 15 {
            switch primaryInputType {
            case .position: fallthrough
            case .velocity: potentialName = primaryInputPosVelNames[element]
                
            case .pressure: fallthrough
            case .forceSingleInput: potentialName = primaryInputSinglePrsFrcNames[element]
                
            case .accelSingleInput: potentialName = primaryInputSingleAccelNames[element]
            
            case .forceDualInputDiff: potentialName = primaryInputDualFrcNames[element]
            case .accelDualInputDiff: potentialName = primaryInputDualAccelNames[element]
                
            default: break
            }
            
            if potentialName != nil && secondaryInputType != .none { potentialName!.name = "Primary \(potentialName!)" }
        }
        
        // Check secondary input type
        else if 23 <= element && element <= 30 {
            let offsetElement = element - 15
            switch secondaryInputType {
            case .forceSingleInput: potentialName = primaryInputSinglePrsFrcNames[offsetElement]
            case .accelSingleInput: potentialName = primaryInputSingleAccelNames[offsetElement]
            case .forceDualInputDiff: potentialName = primaryInputDualFrcNames[offsetElement]
            case .accelDualInputDiff: potentialName = primaryInputDualAccelNames[offsetElement]
            default: break
            }
            
            if potentialName != nil { potentialName!.name = "Secondary \(potentialName!)" }
        }
        return potentialName ?? RMCRegisterInfo(name: "Unknown", unit: .none)
    }
}

// Note: Some of these are exclusive to specific axis types.
// They should probably only appear if the axis is of that type
let commonAxisStatusRegisterNames: [Int: RMCRegisterInfo] = [
    0: RMCRegisterInfo(name: "Status Bits", unit: .bits),
    1: RMCRegisterInfo(name: "Error Bits", unit: .bits),
    2: RMCRegisterInfo(name: "Last Error Number", unit: .none),
    4: RMCRegisterInfo(name: "Read Response", unit: .none),
    
    6: RMCRegisterInfo(name: "Current Control Mode", unit: .none),
    7: RMCRegisterInfo(name: "Next Pos/Vel Control Mode", unit: .none),
    
    12: RMCRegisterInfo(name: "Raw Counts A", unit: .none),
    15: RMCRegisterInfo(name: "Raw Counts B", unit: .none),
    
    33: RMCRegisterInfo(name: "Control Output", unit: .voltageCurrent), // TODO: percent for RMC200
    
    53: RMCRegisterInfo(name: "Target Position", unit: .position),
    54: RMCRegisterInfo(name: "Target Velocity", unit: .velocity),
    55: RMCRegisterInfo(name: "Target Acceleration", unit: .acceleration),
    56: RMCRegisterInfo(name: "Command Position", unit: .position),
    57: RMCRegisterInfo(name: "Command Velocity", unit: .velocity),
    58: RMCRegisterInfo(name: "Target Jerk", unit: .jerk),
    59: RMCRegisterInfo(name: "Cycles", unit: .none),
    60: RMCRegisterInfo(name: "Target Pressure/Force", unit: .pressureForce),
    61: RMCRegisterInfo(name: "Command Pressure/Force", unit: .pressureForce),
    62: RMCRegisterInfo(name: "Cycles (Pressure/Force)", unit: .none),
    
    64: RMCRegisterInfo(name: "Custom Error Bits", unit: .bits),
    65: RMCRegisterInfo(name: "Primary Custom Counts", unit: .none),
    66: RMCRegisterInfo(name: "Secondary Custom Counts", unit: .none)
]

let primaryInputPosVelNames: [Int: RMCRegisterInfo] = [
    8: RMCRegisterInfo(name: "Actual Position", unit: .position),
    9: RMCRegisterInfo(name: "Actual Velocity", unit: .velocity),
    10: RMCRegisterInfo(name: "Actual Acceleration", unit: .acceleration),
    11: RMCRegisterInfo(name: "Counts/Current/Voltage", unit: .voltageCurrent)
]

let primaryInputSinglePrsFrcNames: [Int: RMCRegisterInfo] = [
    8: RMCRegisterInfo(name: "Actual Pressure/Force", unit: .pressureForce),
    9: RMCRegisterInfo(name: "Actual Pressure/Force Rate", unit: .pressureForce),
    11: RMCRegisterInfo(name: "Current/Voltage", unit: .voltageCurrent)
]

let primaryInputSingleAccelNames: [Int: RMCRegisterInfo] = [
    8: RMCRegisterInfo(name: "Actual Acceleration", unit: .acceleration),
    9: RMCRegisterInfo(name: "Actual Jerk", unit: .jerk),
    11: RMCRegisterInfo(name: "Current/Voltage", unit: .voltageCurrent)
]

let primaryInputDualFrcNames: [Int: RMCRegisterInfo] = [
    8: RMCRegisterInfo(name: "Actual Differential Force", unit: .pressureForce),
    9: RMCRegisterInfo(name: "Actual Differential Force Rate", unit: .pressureForceRate),
    10: RMCRegisterInfo(name: "Actual Force A", unit: .pressureForce), // note: only marked as "force" in register reference
    11: RMCRegisterInfo(name: "Voltage A/Current A", unit: .voltageCurrent),
    13: RMCRegisterInfo(name: "Actual Force B", unit: .pressureForce), // note: only marked as "force" in register reference
    14: RMCRegisterInfo(name: "Voltage B/Current B", unit: .voltageCurrent)
]

let primaryInputDualAccelNames: [Int: RMCRegisterInfo] = [
    8: RMCRegisterInfo(name: "Actual Acceleration", unit: .acceleration),
    9: RMCRegisterInfo(name: "Actual Jerk", unit: .jerk),
    10: RMCRegisterInfo(name: "Channel A Acceleration", unit: .pressureForce), // note: only marked as "force" in register reference
    11: RMCRegisterInfo(name: "Voltage A/Current A", unit: .voltageCurrent),
    13: RMCRegisterInfo(name: "Channel B Acceleration", unit: .pressureForce), // note: only marked as "force" in register reference
    14: RMCRegisterInfo(name: "Voltage B/Current B", unit: .voltageCurrent)
]

public struct RMCRegisterInfo {
    public var name: String
    public let unit: RMCRegisterUnit
}

public enum RMCRegisterUnit: String {
    case none = ""
    case bits = "bits"
    case position = "pu"
    case velocity = "pu/s"
    case acceleration = "pu/s2"
    case jerk = "pu/s3"
    case counts = "counts"
    case voltageCurrent = "V or C"
    case pressureForce = "Pr or Fr"
    case pressureForceRate = "Pr/s or Fr/s"
    case percent = "%"
    
    var centerOnZero: Bool {
        switch self {
        case .bits: fallthrough
        case .velocity: fallthrough
        case .acceleration: fallthrough
        case .jerk: fallthrough
        case .voltageCurrent: fallthrough
        case .pressureForce: fallthrough
        case .pressureForceRate: return true
        default: return false
        }
    }
}
