import SwiftSocket
import Foundation

/**
 Holds a connection to an RMC.
 */
public class DMCPConnection {
    public var address: String
    public var port: Int
    
    var client: TCPClient? = nil
    
    /// Whether or not this connection object is currently holding
    /// a live connection to an RMC.
    public private(set) var connected: Bool = false
    
    /// Stores the current transaction ID. Each time a new transaction/request
    /// is sent, this is incremented by one, so that we can uniquely identify
    /// transactions/requests.
    private var currentTransactionID: UInt16 = 0
    
    /// Maps transaction IDs to `DMCPRequest` objects
    private var currentRequests: [UInt16: DMCPRequest] = [:]
    
    /// Executes when disconnected from the RMC.
    public var onDisconnected: () -> Void = { print("Disconnected") }
    
    /// Executes when there is an error reading data, but it cannot be attributed to a specific request
    public var onReadError: () -> Void = { print("Error reading response") }
    
    /// Keeps track of the current RMC type.
    public private(set) var rmcType: RMCType? = nil
    
    private static let productFamilyAddress = DMCPAddress(7,0)
    
    /// Keeps track of current plot configuration data
    public private(set) var plotConfiguration: PlotConfiguration? = PlotConfiguration()
    
    
    private let dq: DispatchQueue
    
    
    /**
     Initializes this connection object with the provided address and port. Note that no actual attempts are made to connect to the RMC until `connect()` is called.
     - Parameters:
        - address: The address of the RMC to connect to. To find this address in RMCTools, right-click on the RMC in the Project View, then click **Connection Path...**. Make sure that the radio button next to **Ethernet** is enabled. The IP address listed below is the RMC's address.
        - port: The port to connect to the RMC on. This is usually `1324`.
     */
    public init(address: String, port: Int = 1324) {
        self.address = address
        self.port = port
        
        dq = DispatchQueue(label: "ThreadSafeDictAccess", attributes: .concurrent)
    }
    
    /**
    Initializes this connection object with an empty address and port of `1324`, the standard for Delta RMCs.
     */
    public init() {
        self.address = ""
        self.port = 1324
        
        dq = DispatchQueue(label: "ThreadSafeDictAccess", attributes: .concurrent)
    }
    
    /**
    Attempts to connect to an RMC.
     - Parameters:
        - onResult: A closure to execute once the connection has either succeeded or failed.
        - _: Describes an error that occurred while attempting to connect to the RMC. When connection is successful, this is `nil`.
     */
    public func connect(_ onResult: @escaping (_ error: Error?) -> Void, onReadError: @escaping () -> Void = {}, onDisconnect: @escaping () -> Void = {}) {
        self.onDisconnected = onDisconnect
        self.client = TCPClient(address: self.address, port: Int32(self.port))
        
        switch client!.connect(timeout: 10) {
        case .success:
            // Set up response capture
            self.connected = true
            DispatchQueue.global(qos: .background).async {
                while self.connected {
                    let available = self.client!.bytesAvailable()
                    
                    if available == nil {
                        print("nil bytes available - connection dropped?")
                        self.connected = false
                    }
                    
                    else if available! > 0 {
                        self.receivedResponse(available!)
                    }
                    
//                    Thread.sleep(forTimeInterval: 0.001)
                }
                
                DispatchQueue.main.async {
                    print("Disconnected")
                    self.onDisconnected()
                }
            }
            
            // Now that we're connected, let's find out what kind of RMC we're connected to
            self.getRMCType { error, type in
                if error != nil {
                    onResult(error)
                    return
                }
                
                self.getCurrentPlotLayout { error, _ in
                    if error != nil && error! as! RMCError != RMCError.unknownRMC {
                        onResult(error)
                        return
                    }
                    
                    self.connected = true
                    onResult(nil)
                }
            }
        case .failure(let error):
            self.connected = false
            onResult(error)
        }
    }
    
    /**
     Disconnects from the RMC and calls the `onDisconnected()` closure, if it has been set.
     */
    public func disconnect() {
        self.connected = false
        client?.close()
        
        DispatchQueue.main.async {
            self.onDisconnected()
        }
    }
    
    func clearClient() {
        print("\(client!.read(Int(client?.bytesAvailable() ?? 0))?.count ?? 0) bytes cleared")
    }
    
    func receivedResponse(_ count: Int32) {
        guard let packetLength = readUInt16() else {
            print("Packet length could not be read. Clearing rest of received packet")
            self.clearClient()
            self.onReadError()
            return
        }
        
        if packetLength < 6 {
            print("Packet not long enough")
            self.onReadError()
            self.clearClient()
            return
        }
        // OK, now we know how many bytes we want to read...
        var bytes: [Byte] = []
        
        while bytes.count < packetLength {
            bytes += client!.read(Int(packetLength)) ?? []
        }
        let restOfData = Data(bytes)
        
        let staticValuesData = restOfData[0...1]
        let transactionIDData = restOfData[2...3]
        let functionCodeResponseData = restOfData[4]
        let rawResponseCodeData = restOfData[5]
        
        var currentDataIndex = 6
        
        func readInt32() -> Int32 {
            let data = restOfData[currentDataIndex..<currentDataIndex + 4] + [0]
            let value = Int32(littleEndian: data.withUnsafeBytes { $0.load(as: Int32.self) })
            currentDataIndex += 4
            return value
        }
        
        func readFloat32() -> Float32 {
            let data = restOfData[currentDataIndex..<currentDataIndex + 4] + [0]
            let value = data.withUnsafeBytes { $0.load(as: Float32.self) }
            
            currentDataIndex += 4
            return value
        }
        
        let staticValues: UInt16 = UInt16(littleEndian: staticValuesData.withUnsafeBytes { $0.load(as: UInt16.self) })
        
        let transactionID: UInt16 = UInt16(littleEndian: transactionIDData.withUnsafeBytes { $0.load(as: UInt16.self) })
        
        let functionCodeResponse: UInt8 = functionCodeResponseData
        
        let rawResponseCode: UInt8 = rawResponseCodeData
        
        let responseCode = DMCPResponseCode(rawValue: rawResponseCode) ?? .unknownResponseCode
        
        dq.async(flags: .barrier) {
            let matchingRequest = self.currentRequests[transactionID]
            
            if matchingRequest != nil {
                self.currentRequests.removeValue(forKey: transactionID)
            }
            
            if matchingRequest is DMCPWriteRequest {
                // emit response for write request
                let writeRequest = matchingRequest as! DMCPWriteRequest
                
                // TODO: ensure that function code response matches!
                if functionCodeResponse != 0x95 {
                    print("Write request response was \(functionCodeResponse), bad.")
                    DispatchQueue.main.async {
                        writeRequest.onResult(DMCPPayloadError.badResponse)
                    }
                    return
                }
                
                DispatchQueue.main.async {
                    writeRequest.onResult(DMCPResponseError.fromCode(responseCode))
                }

            }
            
            else if matchingRequest is DMCPReadRequest {
                // TODO: ensure that function code response matches!
                if functionCodeResponse != 0x94 {
                    print("Read request response was \(functionCodeResponse), bad.")
                    return
                }
                
                // Get the read values
                let readRequest = matchingRequest as! DMCPReadRequest
                let numberOfValues = Int((packetLength - 6) / 4)
                
                let dataTypes = readRequest.readTypes
                var values: [DMCPData] = []
                
                for i in 0..<numberOfValues {
                    switch dataTypes[i] {
                    case .int:
                        values.append(.int(readInt32()))
                    case .real:
                        values.append(.real(readFloat32()))
                    }
                }
                
                
                DispatchQueue.main.async {
                    readRequest.onResult(
                        DMCPResponseError.fromCode(responseCode),
                        values
                    )
                }
            }
        }
    }
    
    /**
     Attempts to send the provided request to the RMC.
     - Parameter request: The request to send. This may be either a `DMCPReadRequest` or a `DMCPWriteRequest`.
     */
    public func sendRequest(_ request: DMCPRequest) {
        dq.async(flags: .barrier) {
            self.currentRequests[self.currentTransactionID] = request

            let data = request.data(transactionID: self.currentTransactionID)
            
            // Increment the current transaction ID, so the next transaction
            // has a different ID
            if self.currentTransactionID == 65535 { self.currentTransactionID = 0 }
            else { self.currentTransactionID += 1 }

            // Send the data, and respond accordingly
            switch self.client?.send(data: data) {
            case .success:
                break
            case .failure(let error):
                if request is DMCPWriteRequest {
                    (request as! DMCPWriteRequest).onResult(error)
                } else if request is DMCPReadRequest {
                    // After a few moments, this starts firing like crazy.
                    // For some reason it can no longer send the read request...
                    // Maybe the buffer is full or something? not sure
                    print("error sending request")
                    (request as! DMCPReadRequest).onResult(error, [])
                }
            case .none:
                break
            }
        }
        

    }
    
    /// Writes the specified `DINT` values to the RMC, starting at the specified address.
    /// - Parameters:
    ///     - address: The address to start writing at.
    ///     - values: An array of `Int`s to write to the RMC as `DINT`s.
    ///     - onResult: A closure to execute once there is a result from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    public func writeInts(_ address: DMCPAddress, values: [Int], onResult: @escaping (_ error: Error?) -> Void) {
        self.writeInts(address.file, address.element, values: values, onResult: onResult)
    }
    
    /// Writes the specified `DINT` values to the RMC, starting at the specified address.
    /// - Parameters:
    ///     - file: The starting address's file part.
    ///     - element: The starting address's element part.
    ///     - values: An array of `Int`s to write to the RMC as `DINT`s.
    ///     - onResult: A closure to execute once there is a result from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    public func writeInts(_ file: UInt16, _ element: UInt16, values: [Int], onResult: @escaping (_ error: Error?) -> Void) {
        let asDMCPData = values.map { i in DMCPData.int(Int32(i)) }
        self.sendRequest(DMCPWriteRequest(file, element, values: asDMCPData, onResult: onResult))
    }
    
    /// Writes the single specified `DINT` value to the RMC at the specified address.
    /// - Parameters:
    ///     - address: The address to write to.
    ///     - value: The `Int` to write to the specified address as a `DINT`.
    ///     - onResult: A closure to execute once there is a result from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    public func writeInt(_ address: DMCPAddress, value: Int, onResult: @escaping (_ error: Error?) -> Void) {
        self.writeInt(address.file, address.element, value: value, onResult: onResult)
    }
    
    /// Writes the single specified `DINT` value to the RMC at the specified address.
    /// - Parameters:
    ///     - file: The address's file part.
    ///     - element: The address's element part.
    ///     - value: The `Int` to write to the specified address as a `DINT`.
    ///     - onResult: A closure to execute once there is a result from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    public func writeInt(_ file: UInt16, _ element: UInt16, value: Int, onResult: @escaping (_ error: Error?) -> Void) {
        self.writeInts(file, element, values: [value], onResult: onResult)
    }
    
    /// Writes the specified `REAL` values to the RMC, starting at the specified address.
    /// - Parameters:
    ///     - address: The address to start writing at.
    ///     - values: An array of `Double`s to write to the RMC as `REAL`s.
    ///     - onResult: A closure to execute once there is a result from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    public func writeReals(_ address: DMCPAddress, values: [Double], onResult: @escaping (_ error: Error?) -> Void) {
        self.writeReals(address.file, address.element, values: values, onResult: onResult)
    }
    
    /// Writes the specified `REAL` values to the RMC, starting at the specified address.
    /// - Parameters:
    ///     - file: The starting address's file part.
    ///     - element: The starting address's element part.
    ///     - values: An array of `Double`s to write to the RMC as `REAL`s.
    ///     - onResult: A closure to execute once there is a result from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    public func writeReals(_ file: UInt16, _ element: UInt16, values: [Double], onResult: @escaping (_ error: Error?) -> Void) {
        let asDMCPData = values.map { i in DMCPData.real(Float32(i)) }
        self.sendRequest(DMCPWriteRequest(file, element, values: asDMCPData, onResult: onResult))
    }
    
    /// Writes the single specified `REAL` value to the RMC at the specified address.
    /// - Parameters:
    ///     - address: The address to write to.
    ///     - value: The `Double` to write to the specified address as a `REAL`.
    ///     - onResult: A closure to execute once there is a result from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    public func writeReal(_ address: DMCPAddress, value: Double, onResult: @escaping (_ error: Error?) -> Void) {
        self.writeReal(address.file, address.element, value: value, onResult: onResult)
    }
    
    /// Writes the single specified `REAL` value to the RMC at the specified address.
    /// - Parameters:
    ///     - file: The address's file part.
    ///     - element: The address's element part.
    ///     - value: The `Double` to write to the specified address as a `REAL`.
    ///     - onResult: A closure to execute once there is a result from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    public func writeReal(_ file: UInt16, _ element: UInt16, value: Double, onResult: @escaping (_ error: Error?) -> Void) {
        self.writeReals(file, element, values: [value], onResult: onResult)
    }
    
    /// Writes the array of `DMCPData` items to the RMC, starting at the specified address.
    /// - Parameters:
    ///     - address: The address to start writing to.
    ///     - values: An array of values to write to the RMC.
    ///     - onResult: A closure to execute once there is a response from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    public func writeValues(_ address: DMCPAddress, values: [DMCPData], onResult: @escaping (_ error: Error?) -> Void) {
        self.writeValues(address.file, address.element, values: values, onResult: onResult)
    }
    
    /// Writes the array of `DMCPData` items to the RMC, starting at the specified address.
    /// - Parameters:
    ///     - file: The address's file part.
    ///     - element: The address's element part.
    ///     - values: An array of values to write to the RMC.
    ///     - onResult: A closure to execute once there is a response from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    public func writeValues(_ file: UInt16, _ element: UInt16, values: [DMCPData], onResult: @escaping (_ error: Error?) -> Void) {
        self.sendRequest(DMCPWriteRequest(file, element, values: values, onResult: onResult))
    }
    
    /// Reads `DINT` values from the RMC, starting at the specified address.
    /// - Parameters:
    ///     - address: The address to start reading from.
    ///     - count: The number of values to read.
    ///     - onResult: A closure to execute once there is a result from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    ///     - values: The array of values retrieved. If an error occurred, this is empty.
    public func readInts(_ address: DMCPAddress, count: Int = 1, onResult: @escaping (_ error: Error?, _ values: [Int?]) -> Void) {
        self.readInts(address.file, address.element, count: count, onResult: onResult)
    }
    
    /// Reads `DINT` values from the RMC, starting at the specified address.
    /// - Parameters:
    ///     - file: The starting address's file part.
    ///     - element: The starting address's element part.
    ///     - count: The number of values to read.
    ///     - onResult: A closure to execute once there is a result from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    ///     - values: The array of values retrieved. If an error occurred, this is empty.
    public func readInts(_ file: UInt16, _ element: UInt16, count: Int = 1, onResult: @escaping (_ error: Error?, _ values: [Int?]) -> Void) {
        self.sendRequest(DMCPReadRequest(file, element, readTypes: [DMCPData.DataType].init(repeating: .int, count: count)) { error, rawResult in
            let asInts = rawResult.map { i -> Int? in
                if case .int(let val) = i {
                    return val != nil ? Int(val!) : nil
                } else if case .real(let val) = i {
                    return val != nil ? Int(val!) : nil
                } else {
                    return 0
                }
            }
            
            onResult(error, asInts)
        })
    }
    
    /// Reads a single `DINT` value from the RMC at the specified address.
    /// - Parameters:
    ///     - address: The address to read from.
    ///     - onResult: A closure to execute once there is a result from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    ///     - value: The value retrieved. If an error occurred, this is `nil`.
    public func readInt(_ address: DMCPAddress, onResult: @escaping (_ error: Error?, _ value: Int?) -> Void) {
        self.readInt(address.file, address.element, onResult: onResult)
    }
    
    /// Reads a single `DINT` value from the RMC at the specified address.
    /// - Parameters:
    ///     - file: The address's file part.
    ///     - element: The address's element part.
    ///     - onResult: A closure to execute once there is a result from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    ///     - value: The value retrieved. If an error occurred, this is `nil`.
    public func readInt(_ file: UInt16, _ element: UInt16, onResult: @escaping (_ error: Error?, _ value: Int?) -> Void) {
        self.readInts(file, element) { error, array in
            onResult(error, !array.isEmpty ? array[0] : nil)
        }
    }
    
    /// Reads `REAL` values from the RMC, starting at the specified address.
    /// - Parameters:
    ///     - address: The address to start reading from.
    ///     - count: The number of values to read.
    ///     - onResult: A closure to execute once there is a result from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    ///     - values: The array of values retrieved. If an error occurred, this is empty.
    public func readReals(_ address: DMCPAddress, count: Int = 1, onResult: @escaping (_ error: Error?, _ values: [Double?]) -> Void) {
        self.readReals(address.file, address.element, count: count, onResult: onResult)
    }

    /// Reads `REAL` values from the RMC, starting at the specified address.
    /// - Parameters:
    ///     - file: The starting address's file part.
    ///     - element: The starting address's element part.
    ///     - count: The number of values to read.
    ///     - onResult: A closure to execute once there is a result from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    ///     - values: The array of values retrieved. If an error occurred, this is empty.
    public func readReals(_ file: UInt16, _ element: UInt16, count: Int = 1, onResult: @escaping (_ error: Error?, _ values: [Double?]) -> Void) {
        self.sendRequest(DMCPReadRequest(file, element, readTypes: [DMCPData.DataType].init(repeating: .real, count: count)) { error, rawResult in
            let asReals = rawResult.map { i -> Double? in
                if case .real(let val) = i { return val != nil ? Double(val!) : nil }
                else if case .int(let val) = i { return val != nil ? Double(val!) : nil }
                else { return 0 }
            }
            
            onResult(error, asReals)
        })
    }
    
    /// Reads a single `REAL` value from the RMC at the specified address.
    /// - Parameters:
    ///     - address: The address to read from.
    ///     - onResult: A closure to execute once there is a result from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    ///     - value: The value retrieved. If an error occurred, this is `nil`.
    public func readReal(_ address: DMCPAddress, onResult: @escaping (_ error: Error?, _ value: Double?) -> Void) {
        self.readReal(address.file, address.element, onResult: onResult)
    }
    
    /// Reads a single `REAL` value from the RMC at the specified address.
    /// - Parameters:
    ///     - file: The address's file part.
    ///     - element: The address's element part.
    ///     - onResult: A closure to execute once there is a result from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    ///     - value: The value retrieved. If an error occurred, this is `nil`.
    public func readReal(_ file: UInt16, _ element: UInt16, onResult: @escaping (_ error: Error?, _ value: Double?) -> Void) {
        self.readReals(file, element) { error, array in
            onResult(error, !array.isEmpty ? array[0] : nil)
        }
    }
    
    /// Reads a string from the RMC at the specified address.
    /// - Parameters:
    ///     - address: The address to read from.
    ///     - maxLength: The maximum number of characters to read. If a null byte (`\0`) is encountered, the string will terminate immediately.
    ///     - onResult: A closure to execute once there is a result from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    ///     - value: The value retrieved. If an error occurred, this is `nil`.
    public func readString(_ address: DMCPAddress, maxLength: Int, onResult: @escaping (_ error: Error?, _ value: String?) -> Void) {
        self.readString(address.file, address.element, maxLength: maxLength, onResult: onResult)
    }
    
    /// Reads a string from the RMC at the specified address.
    /// - Parameters:
    ///     - file: The address's file part.
    ///     - element: The address's element part.
    ///     - maxLength: The maximum number of characters to read. If a null byte (`\0`) is encountered, the string will terminate immediately.
    ///     - onResult: A closure to execute once there is a result from this request.
    ///     - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
    ///     - value: The value retrieved. If an error occurred, this is `nil`.
    public func readString(_ file: UInt16, _ element: UInt16, maxLength: Int, onResult: @escaping (_ error: Error?, _ value: String?) -> Void) {
        self.readInts(file, element, count: maxLength) { error, intArray in
            var output = ""
            
            for item in intArray {
                let newPieceAsData = Swift.withUnsafeBytes(of: item) { Data($0) }
                let a = newPieceAsData.withUnsafeBytes { $0.load(as: (UInt16, UInt16).self) }
                let c1 = Character(UnicodeScalar(a.1)!)
                let c2 = Character(UnicodeScalar(a.0)!)
                
                if c1 == "\0" {
                    break
                } else if c2 == "\0" {
                    output += [c1]
                    break
                }
                output += [c1, c2]
            }
            onResult(error, !intArray.isEmpty ? output : nil)
        }
    }
    
    /**
     Reads the name of a specific axis.
     - Parameters:
        - axisIndex: The index of the axis to retrieve the name for.
        - onResult: A closure to execute once there is a result from this request.
        - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
        - value: The axis name. If an error occurred, this is `nil`.
     */
    public func getAxisName(_ axisIndex: Int, onResult: @escaping (_ error: Error?, _ value: String?) -> Void) {
        // Determine what address we need to use
        let address = self.rmcType?.toRMCClass()?.axisNameAddr(axisIndex)
        if address == nil { onResult(RMCError.unknownRMC, nil); return }
        self.readString(address!, maxLength: 16, onResult: onResult)
    }
    
    /**
     Reads the name of the controller.
     - Parameters:
        - onResult: A closure to execute once there is a result from this request.
        - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
        - value: The controller name. If an error occurred, this is `nil`.
     */
    public func getControllerName(_ onResult: @escaping (_ error: Error?, _ value: String?) -> Void) {
        let address = self.rmcType?.toRMCClass()?.controllerNameAddr
        if address == nil { onResult(RMCError.unknownRMC, nil); return }
        self.readString(address!, maxLength: 16, onResult: onResult)
    }
    
    /**
     Reads the 8-digit serial number of the controller.
     - Parameters:
        - onResult: A closure to execute once there is a result from this request.
        - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
        - value: The controller's serial number. If an error occurred, this is `nil`.
     */
    public func getSerialNumber(_ onResult: @escaping (_ error: Error?, _ value: Int?) -> Void) {
        let address = self.rmcType?.toRMCClass()?.serialNumberAddr
        if address == nil { onResult(RMCError.unknownRMC, nil); return }
        self.readInt(address!, onResult: onResult)
    }
    
    /**
    Reads the RMC's product family (RMC75, RMC150, or RMC200).
     When successful, this function updates the value stored in `rmcType`.
    - Parameters:
        - onResult: A closure to execute once there is a result from this request.
        - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
        - value: The product family of the RMC. If an error occurred, this is `nil`.
    */
    public func getRMCType(_ onResult: @escaping (_ error: Error?, _ value: RMCType?) -> Void = {_,_ in}) {
        self.readReal(DMCPConnection.productFamilyAddress) { error, result in
            var result = result
            if result != nil {
                
                // The RMC200 gives its product ID as an int, not a float, so
                // we have to convert it
                if result == 5.605193857299268e-45 {
                    result = 4.0
                }
                self.rmcType = RMCType(rawValue: result!)
            }
            onResult(error, self.rmcType)
        }
    }
    
    /**
     Reads the type information about a specific axis.
     - Parameters:
        - axisIndex: The index of the axis to retrieve information for.
        - onResult: A closure to execute once there is a result from this request.
        - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
        - value: The axis type, primary input type, and secondary input type as a 3-tuple. If an error occurred, this is `nil`.
     */
    public func getAxisType(forAxis axisIndex: Int, _ onResult: @escaping (_ error: Error?, _ value: (AxisType, InputType, InputType)?) -> Void) {
        var baseAddress = self.rmcType?.toRMCClass()?.currentAxisDefinitionsAddr
        if baseAddress == nil { onResult(RMCError.unknownRMC, nil); return }
        baseAddress?.element = UInt16(axisIndex * 4)
        
        self.readInt(baseAddress!) { error, result in
            if error != nil { onResult(error, nil); return }
            
            let axisType = AxisType(rawValue: result! & 0xFF)!
            let primaryInputType = InputType(rawValue: (result! >> 16) & 0xFF)!
            let secondaryInputType = InputType(rawValue: (result! >> 24) & 0xFF)!
            onResult(nil, (axisType, primaryInputType, secondaryInputType))
        }
    }
    
    /**
    Finds a human-readable label for a specific address, if possible.
    - Parameters:
        - address: The address to find a label for.
        - onResult: A closure to execute once there is a result from this request.
        - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
        - value: The human-readable label. If an error occurred, this will be as complete of a string as possible.
     */
    public func getAddressLabel(for address: DMCPAddress, _ onResult: @escaping (_ error: Error?, _ value: RMCRegisterInfo?) -> Void) {
        // Calculate the axis
        let axisNumber = address.file - 8
        
        // If this isn't an axis number, then return nil
        let numberOfAxes = self.rmcType?.toRMCClass()?.numberOfAxes
        guard let numberOfAxes = numberOfAxes else { onResult(RMCError.unknownRMC, nil); return }
        
        if !(0..<numberOfAxes).contains(Int(axisNumber)) {
            print("\(address) isn't for an axis!")
            onResult(nil, RMCRegisterInfo(name: "Unlabeled Register", unit: .none))
            return // TODO: make this throw an error or something
        }
        
        // This is an axis! Let's get axis info
        self.getAxisType(forAxis: Int(axisNumber)) { error, axisInfo in
            if error != nil {
                print("Error occurred trying to get axis info: \(error!)")
                onResult(error, RMCRegisterInfo(name: "Axis \(axisNumber) Register", unit: .none))
                return
            }
            
            let (axisType, primaryInputType, secondaryInputType) = axisInfo!
            
            // We now have the axis information, so let's get the name from it
            let displayName = self.rmcType?.toRMCClass()?.getDisplayName(Int(address.element), axisType: axisType, primaryInputType: primaryInputType, secondaryInputType: secondaryInputType) ?? RMCRegisterInfo(name: "Unlabeled Register", unit: .none)
            
            // Finally let's add the axis name
            self.getAxisName(Int(axisNumber)) { error, name in
                if error != nil {
                    print("Error occurred trying to get axis name: \(error!)")
                    onResult(error, displayName)
                    return
                }
                
                let fullName = "\(name!) \(displayName.name)"
                onResult(nil, RMCRegisterInfo(name: fullName, unit: displayName.unit))
            }
        }
    }
    
    /**
     Reads the RMC's current plot layout (the number of plots currently allocated, and the number of sample sets available per plot).
     When successful, this function updates the value stored in `plotConfiguration`.
     - Parameters:
        - onResult: A closure to execute once there is a result from this request.
        - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
        - value: A structure containing information about the RMC's current plot configuration. If an error occurred, this is `nil`.
     */
    public func getCurrentPlotLayout(_ onResult: @escaping (_ error: Error?, _ value: PlotConfiguration?) -> Void = {_,_ in}) {
        let address = self.rmcType?.toRMCClass()?.currentPlotLayoutAddr
        guard let address = address else { onResult(RMCError.unknownRMC, nil); return }
        
        self.readReal(address) { error, rawValue in
            if rawValue == nil {
                onResult(error, nil)
                return
            }
            
            let asInt = Int(rawValue!)
            
            // Lower 8 bits represent number of plots
            let numberOfPlots = asInt & 255
            let sampleSetsPerPlot = (asInt >> 8) & 255
            
            var newPlotConfig = PlotConfiguration()
            newPlotConfig.numberOfPlots = numberOfPlots
            newPlotConfig.sampleSetsPerPlot = sampleSetsPerPlot
            
            self.plotConfiguration = newPlotConfig
            
            onResult(error, self.plotConfiguration)
        }
    }
    
    /**
     Reads the maximum number of plots the RMC may allocate.
     - Parameters:
        - onResult: A closure to execute once there is a result from this request.
        - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
        - value: The maximum number of plots the RMC may allocate. If an error occurred, this is `nil`.
     */
    public func getMaximumPlots(_ onResult: @escaping (_ error: Error?, _ value: Int?) -> Void) {
        guard let address = self.rmcType?.toRMCClass()?.maximumPlotsAddr else { onResult(RMCError.unknownRMC, nil); return }
        self.readReal(address, onResult: {error, i in onResult(error, i != nil ? Int(i!) : nil)})
    }
    
    /**
     Reads the maximum number of data items the RMC may capture per plot.
     - Parameters:
        - onResult: A closure to execute once there is a result from this request.
        - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
        - value: The maximum number of data items the RMC may capture per plot. If an error occurred, this is `nil`.
     */
    public func getPlotMaximumDataItems(_ onResult: @escaping (_ error: Error?, _ value: Int?) -> Void) {
        guard let address = self.rmcType?.toRMCClass()?.maximumPlotsDataItemsAddr else { onResult(RMCError.unknownRMC, nil); return }
        self.readReal(address, onResult: {error, i in onResult(error, i != nil ? Int(i!) : nil)})
    }
    
    /**
     Issues a command to the RMC.
     
     For a list of valid commands, see ["RMC Commands" in the RMCTools Help documentation](https://deltamotion.com/support/webhelp/rmctools/Command_Reference/Command_List.htm).
     - Warning: The `onResult` closure passing a `nil` value for `error` does *not* mean that the command was accepted by the RMC.
     - Parameters:
        - command: The command to issue.
        - axis: The axis to which the command should be issued.
        - onResult: A closure to execute once there is a result from this request.
        - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
     */
    public func issueCommand(_ command: RMCCommand, toAxis axis: Int, _ onResult: @escaping (_ error: Error?) -> Void = {_ in}) {
        guard let address = self.rmcType?.toRMCClass()?.axisCommandAddr(axis) else { onResult(RMCError.unknownRMC); return }
        self.writeReals(address, values: command.array(), onResult: onResult)
    }
    
    /**
     Reads the number of samples a given plot is configured to capture.
     
     - Parameters:
        - plot: The ID of the plot.
        - onResult: A closure to execute once there is a result from this request.
        - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
        - value: The number of samples this plot is configured to capture. If an error occurred, this is `nil`.
     */
    public func getPlotSamples(forPlot plot: Int, _ onResult: @escaping (_ error: Error?, _ value: Int?) -> Void) {
        guard let address = self.rmcType?.toRMCClass()?.plotSamplesAddr(plot) else { onResult(RMCError.unknownRMC, nil); return }
        self.readReal(address, onResult: { error, i in onResult(error, i != nil ? Int(i!) : nil)})
    }
    
    /**
     Reads the amount of time between sample captures for a given plot.
     
     - Parameters:
        - plot: The ID of the plot.
        - onResult: A closure to execute once there is a result from this request.
        - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
        - value: The amount of time between sample captures for this plot. If an error occurred, this is `nil`.
     */
    public func getPlotSamplePeriod(forPlot plot: Int, _ onResult: @escaping (_ error: Error?, _ value: Double?) -> Void) {
        guard let address = self.rmcType?.toRMCClass()?.plotSamplePeriodAddr(plot) else { onResult(RMCError.unknownRMC, nil); return }
        self.readReal(address, onResult: { error, i in onResult(error, i)})
    }
    
    /**
     Reads the axis that owns a given plot.
     
     - Parameters:
        - plot: The ID of the plot.
        - onResult: A closure to execute once there is a result from this request.
        - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
        - value: The number of the axis that owns this plot. If an error occurred, this is `nil`.
     */
    public func getPlotAxisOwner(forPlot plot: Int, _ onResult: @escaping (_ error: Error?, _ value: Int?) -> Void) {
        guard let address = self.rmcType?.toRMCClass()?.plotAxisOwnerAddr(plot) else { onResult(RMCError.unknownRMC, nil); return }
        self.readReal(address, onResult: { error, i in onResult(error, i != nil ? Int(i!) : nil)})
    }
    
    /**
     Reads all of the addresses that a given plot is tracking.
     
     - Parameters:
        - plot: The ID of the plot.
        - onResult: A closure to execute once there is a result from this request.
        - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
        - values: An ordered array of `DMCPAddress` structures, representing the addresses that this plot is tracking. If an error occurred, this is an empty array.
     */
    public func getPlotAddresses(forPlot plot: Int, _ onResult: @escaping (_ error: Error?, _ values: [DMCPAddress]) -> Void) {
        guard let address = self.rmcType?.toRMCClass()?.plotDataSetsAddressesAddr(plot) else { onResult(RMCError.unknownRMC, []); return }
        
        self.readReals(address, count: 16) { error, values in
            let endPosition = (values.firstIndex(of: 0.0) ?? values.count) - 1
            print(values)
            onResult(nil, values[...endPosition].map { DMCPAddress(fromWord: Int($0!)) })
        }
    }
    
    /**
     Reads a single address that a given plot is tracking.

     - Parameters:
        - plot: The ID of the plot.
        - onResult: A closure to execute once there is a result from this request.
        - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
        - value: A single address that this plot is tracking. If an error occurred, this is  `nil`.
     */
    public func getPlotAddress(forPlot plot: Int, atIndex index: Int, _ onResult: @escaping (_ error: Error?, _ value: DMCPAddress?) -> Void) {
        guard var address = self.rmcType?.toRMCClass()?.plotDataSetsAddressesAddr(plot) else { onResult(RMCError.unknownRMC, nil); return }
        address.element += UInt16(index)
        self.readReal(address) { error, val in
            if val == nil { onResult(error, nil); return }
            onResult(nil, DMCPAddress(fromWord: Int(val!)))
        }
    }
    
    /**
     Reads the number of tasks that the RMC has allocated.
     
     - Parameters:
        - onResult: A closure to execute once there is a result from this request.
        - error: Describes an error that occurred while attempting to communicate with the RMC. If the operation was successful, this is `nil`.
        - value: The number of tasks the RMC has allocated. If an error occurred, this is `nil`.
     */
    public func getNumberOfAllocatedTasks(_ onResult: @escaping (_ error: Error?, _ value: Int?) -> Void) {
        guard let address = self.rmcType?.toRMCClass()?.numberOfAllocatedTasksAddr else { onResult(RMCError.unknownRMC, nil); return }
        self.readReal(address) { error, value in
            onResult(error, value != nil ? Int(value!) : nil)
        }
    }
    
    func readUInt16() -> UInt16? {
        guard let byteArray = client?.read(2) else { return nil }
        let data = Data(byteArray)
        let value: UInt16 = UInt16(littleEndian: data.withUnsafeBytes { $0.load(as: UInt16.self) })
        return value
    }
    
    func readUInt8() -> UInt8? {
        guard let byteArray = client?.read(1) else { return nil }
        let data = Data(byteArray)
        let value = UInt8(littleEndian: data.withUnsafeBytes { $0.load(as: UInt8.self) })
        return value
    }
    
    func readInt32() -> Int32? {
        guard let byteArray = client?.read(4) else { return nil }
        let data = Data(byteArray)
        let value = Int32(littleEndian: data.withUnsafeBytes { $0.load(as: Int32.self) })
        return value
    }
    
    // TODO: make this work asynchronously, building up bytes until it has 4 of them
    func readFloat32() -> Float32? {
        guard let byteArray = client?.read(4) else { return nil }
        let data = Data(byteArray)
        let value = data.withUnsafeBytes { $0.load(as: Float32.self) }
        return value
    }
}
